#include <credits.h>

namespace sdrpp_credits {
    const char* contributors[] = {
        "Aang23",
        "Alexsey Shestacov",
        "Aosync",
        "Benjamin Kyd",
        "Benjamin Vernoux",
        "Cropinghigh",
        "Fred F4EED",
        "Howard0su",
        "John Donkersley",
        "Joshua Kimsey",
        "Manawyrm",
        "Martin Hauke",
        "Marvin Sinister",
        "Maxime Biette",
        "Paulo Matias",
        "Raov",
        "Cam K.",
        "Shuyuan Liu",
        "Syne Ardwin (WI9SYN)",
        "Szymon Zakrent",
        "Youssef Touil",
        "Zimm"
    };

    const char* libraries[] = {
        "Dear ImGui (ocornut)",
        "fftw3 (fftw.org)",
        "glfw (Camilla Löwy)",
        "json (nlohmann)",
        "Portable File Dialogs"
    };

    const char* hardwareDonators[] = {
        "Aaronia AG",
        "Airspy",
        "Alex 4Z5LV",
        "Analog Devices",
        "CaribouLabs",
        "Deepace",
        "Ettus Research",
        "Harogic",
        "Howard Su",
        "MicroPhase",
        "Microtelecom",
        "MyriadRF",
        "Nuand",
        "RFNM",
        "RFspace",
        "RigExpert",
        "RTL-SDRblog",
        "SDRplay"
    };

    const char* patrons[] = {
        "Bob Logan",
        "Christian Häusler",
        "Croccydile",
        "Dale L Puckett (K0HYD)",
        "Daniele D'Agnelli",
        "David Taylor (GM8ARV)",
        "D. Jones",
        "Dexruus",
        "EB3FRN",
        "Eric Johnson",
        "Ernest Murphy (NH7L)",
        "Flinger Films",
        "Frank Werner (HB9FXQ)",
        "gringogrigio",
        "Jandro",
        "Jeff Moe",
        "Joe Cupano",
        "KD1SQ",
        "Kezza",
        "Krys Kamieniecki",
        "Lee Donaghy",
        "Lee (KD1SQ)",
        ".lozenge. (Hank Hill)",
        "Martin Herren (HB9FXX)",
        "NeoVilsonWong",
        "Nitin (VU2JEK)",
        "ON4MU",
        "Passion-Radio.com",
        "Paul Maine",
        "Peter Betz",
        "Scanner School",
        "Scott Palmer",
        "SignalsEverywhere",
        "Syne Ardwin (WI9SYN)",
        "W4IPA",
        "William Arcand (W1WRA)",
        "William Pitchford",
        "Yves Rougy",
        "Zipper"
    };

    const int contributorCount = sizeof(contributors) / sizeof(char*);
    const int libraryCount = sizeof(libraries) / sizeof(char*);
    const int hardwareDonatorCount = sizeof(hardwareDonators) / sizeof(char*);
    const int patronCount = sizeof(patrons) / sizeof(char*);
}
