#pragma once

enum {
    RECORDER_IFACE_CMD_GET_MODE,
    RECORDER_IFACE_CMD_SET_MODE,
    RECORDER_IFACE_CMD_START,
    RECORDER_IFACE_CMD_STOP
};

enum {
    RECORDER_MODE_BASEBAND,
    RECORDER_MODE_AUDIO
};